try:
    from socket import inet_pton
except ImportError:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:  # pragma: no cover
        pass
    else:
        # based on https://gist.github.com/nnemkin/4966028
        # this code only applies on Windows Python 2.7
        import ctypes
        import socket

        class SockAddr(ctypes.Structure):
            _fields_ = [
                ("sa_family", ctypes.c_short),
                ("__pad1", ctypes.c_ushort),
                ("ipv4_addr", ctypes.c_byte * 4),
                ("ipv6_addr", ctypes.c_byte * 16),
                ("__pad2", ctypes.c_ulong),
            ]

        WSAStringToAddressA = ctypes.windll.ws2_32.WSAStringToAddressA
        WSAAddressToStringA = ctypes.windll.ws2_32.WSAAddressToStringA

        def inet_pton(address_family, ip_string):
            # type: (int, str) -> bytes
            addr = SockAddr()
            ip_string_bytes = ip_string.encode("ascii")
            addr.sa_family = address_family
            addr_size = ctypes.c_int(ctypes.sizeof(addr))

            try:
                attribute, size = {
                    socket.AF_INET: ("ipv4_addr", 4),
                    socket.AF_INET6: ("ipv6_addr", 16),
                }[address_family]
            except KeyError:
                raise socket.error("unknown address family")

            if (
                WSAStringToAddressA(
                    ip_string_bytes,
                    address_family,
                    None,
                    ctypes.byref(addr),
                    ctypes.byref(addr_size),
                )
                != 0
            ):
                raise socket.error(ctypes.FormatError())

            return ctypes.string_at(getattr(addr, attribute), size)



from typing import Any, Callable, Optional, Type
from unittest import TestCase


class HyperlinkTestCase(TestCase):
    """This type mostly exists to provide a backwards-compatible
    assertRaises method for Python 2.6 testing.
    """

    def assertRaises(  # type: ignore[override]
        self,
        expected_exception,  # type: Type[BaseException]
        callableObj=None,  # type: Optional[Callable[..., Any]]
        *args,  # type: Any
        **kwargs  # type: Any
    ):
        # type: (...) -> Any
        """Fail unless an exception of class expected_exception is raised
        by callableObj when invoked with arguments args and keyword
        arguments kwargs. If a different type of exception is
        raised, it will not be caught, and the test case will be
        deemed to have suffered an error, exactly as for an
        unexpected exception.

        If called with callableObj omitted or None, will return a
        context object used like this::

             with self.assertRaises(SomeException):
                 do_something()

        The context manager keeps a reference to the exception as
        the 'exception' attribute. This allows you to inspect the
        exception after the assertion::

            with self.assertRaises(SomeException) as cm:
                do_something()
            the_exception = cm.exception
            self.assertEqual(the_exception.error_code, 3)
        """
        context = _AssertRaisesContext(expected_exception, self)
        if callableObj is None:
            return context
        with context:
            callableObj(*args, **kwargs)


class _AssertRaisesContext(object):
    "A context manager used to implement HyperlinkTestCase.assertRaises."

    def __init__(self, expected, test_case):
        # type: (Type[BaseException], TestCase) -> None
        self.expected = expected
        self.failureException = test_case.failureException

    def __enter__(self):
        # type: () -> "_AssertRaisesContext"
        return self

    def __exit__(self, exc_type, exc_value, tb):
        # type: (Optional[Type[BaseException]], Any, Any) -> bool
        if exc_type is None:
            exc_name = self.expected.__name__
            raise self.failureException("%s not raised" % (exc_name,))
        if not issubclass(exc_type, self.expected):
            # let unexpected exceptions pass through
            return False
        self.exception = exc_value  # store for later retrieval
        return True



"""
Tests for hyperlink.test.common
"""
from typing import Any
from unittest import TestCase
from .common import HyperlinkTestCase


class _ExpectedException(Exception):
    """An exception used to test HyperlinkTestCase.assertRaises."""


class _UnexpectedException(Exception):
    """An exception used to test HyperlinkTestCase.assertRaises."""


class TestHyperlink(TestCase):
    """Tests for HyperlinkTestCase"""

    def setUp(self):
        # type: () -> None
        self.hyperlink_test = HyperlinkTestCase("run")

    def test_assertRaisesWithCallable(self):
        # type: () -> None
        """HyperlinkTestCase.assertRaises does not raise an AssertionError
        when given a callable that, when called with the provided
        arguments, raises the expected exception.

        """
        called_with = []

        def raisesExpected(*args, **kwargs):
            # type: (Any, Any) -> None
            called_with.append((args, kwargs))
            raise _ExpectedException

        self.hyperlink_test.assertRaises(
            _ExpectedException, raisesExpected, 1, keyword=True
        )
        self.assertEqual(called_with, [((1,), {"keyword": True})])

    def test_assertRaisesWithCallableUnexpectedException(self):
        # type: () -> None
        """When given a callable that raises an unexpected exception,
        HyperlinkTestCase.assertRaises raises that exception.

        """

        def doesNotRaiseExpected(*args, **kwargs):
            # type: (Any, Any) -> None
            raise _UnexpectedException

        try:
            self.hyperlink_test.assertRaises(
                _ExpectedException, doesNotRaiseExpected
            )
        except _UnexpectedException:
            pass

    def test_assertRaisesWithCallableDoesNotRaise(self):
        # type: () -> None
        """HyperlinkTestCase.assertRaises raises an AssertionError when given
        a callable that, when called, does not raise any exception.

        """

        def doesNotRaise(*args, **kwargs):
            # type: (Any, Any) -> None
            pass

        try:
            self.hyperlink_test.assertRaises(_ExpectedException, doesNotRaise)
        except AssertionError:
            pass

    def test_assertRaisesContextManager(self):
        # type: () -> None
        """HyperlinkTestCase.assertRaises does not raise an AssertionError
        when used as a context manager with a suite that raises the
        expected exception.  The context manager stores the exception
        instance under its `exception` instance variable.

        """
        with self.hyperlink_test.assertRaises(_ExpectedException) as cm:
            raise _ExpectedException

        self.assertTrue(  # type: ignore[unreachable]
            isinstance(cm.exception, _ExpectedException)
        )

    def test_assertRaisesContextManagerUnexpectedException(self):
        # type: () -> None
        """When used as a context manager with a block that raises an
        unexpected exception, HyperlinkTestCase.assertRaises raises
        that unexpected exception.

        """
        try:
            with self.hyperlink_test.assertRaises(_ExpectedException):
                raise _UnexpectedException
        except _UnexpectedException:
            pass

    def test_assertRaisesContextManagerDoesNotRaise(self):
        # type: () -> None
        """HyperlinkTestcase.assertRaises raises an AssertionError when used
        as a context manager with a block that does not raise any
        exception.

        """
        try:
            with self.hyperlink_test.assertRaises(_ExpectedException):
                pass
        except AssertionError:
            pass



# -*- coding: utf-8 -*-
"""
Tests for hyperlink.hypothesis.
"""

try:
    import hypothesis

    del hypothesis
except ImportError:
    pass
else:
    from string import digits
    from typing import Sequence, Text

    try:
        from unittest.mock import patch
    except ImportError:
        from mock import patch  # type: ignore[misc]

    from hypothesis import given, settings
    from hypothesis.strategies import SearchStrategy, data

    from idna import IDNAError, check_label, encode as idna_encode

    from .common import HyperlinkTestCase
    from .. import DecodedURL, EncodedURL
    from ..hypothesis import (
        DrawCallable,
        composite,
        decoded_urls,
        encoded_urls,
        hostname_labels,
        hostnames,
        idna_text,
        paths,
        port_numbers,
    )

    class TestHypothesisStrategies(HyperlinkTestCase):
        """
        Tests for hyperlink.hypothesis.
        """

        @given(idna_text())
        def test_idna_text_valid(self, text):
            # type: (Text) -> None
            """
            idna_text() generates IDNA-encodable text.
            """
            try:
                idna_encode(text)
            except IDNAError:  # pragma: no cover
                raise AssertionError("Invalid IDNA text: {!r}".format(text))

        @given(data())
        def test_idna_text_min_max(self, data):
            # type: (SearchStrategy) -> None
            """
            idna_text() raises AssertionError if min_size is < 1.
            """
            self.assertRaises(AssertionError, data.draw, idna_text(min_size=0))
            self.assertRaises(AssertionError, data.draw, idna_text(max_size=0))

        @given(port_numbers())
        def test_port_numbers_bounds(self, port):
            # type: (int) -> None
            """
            port_numbers() generates integers between 1 and 65535, inclusive.
            """
            self.assertGreaterEqual(port, 1)
            self.assertLessEqual(port, 65535)

        @given(port_numbers(allow_zero=True))
        def test_port_numbers_bounds_allow_zero(self, port):
            # type: (int) -> None
            """
            port_numbers(allow_zero=True) generates integers between 0 and
            65535, inclusive.
            """
            self.assertGreaterEqual(port, 0)
            self.assertLessEqual(port, 65535)

        @given(hostname_labels())
        def test_hostname_labels_valid_idn(self, label):
            # type: (Text) -> None
            """
            hostname_labels() generates IDN host name labels.
            """
            try:
                check_label(label)
                idna_encode(label)
            except UnicodeError:  # pragma: no cover
                raise AssertionError("Invalid IDN label: {!r}".format(label))

        @given(data())
        @settings(max_examples=10)
        def test_hostname_labels_long_idn_punycode(self, data):
            # type: (SearchStrategy) -> None
            """
            hostname_labels() handles case where idna_text() generates text
            that encoded to punycode ends up as longer than allowed.
            """

            @composite
            def mock_idna_text(draw, min_size, max_size):
                # type: (DrawCallable, int, int) -> Text
                # We want a string that does not exceed max_size, but when
                # encoded to punycode, does exceed max_size.
                # So use a unicode character that is larger when encoded,
                # "á" being a great example, and use it max_size times, which
                # will be max_size * 3 in size when encoded.
                return u"\N{LATIN SMALL LETTER A WITH ACUTE}" * max_size

            with patch("hyperlink.hypothesis.idna_text", mock_idna_text):
                label = data.draw(hostname_labels())
                try:
                    check_label(label)
                    idna_encode(label)
                except UnicodeError:  # pragma: no cover
                    raise AssertionError(
                        "Invalid IDN label: {!r}".format(label)
                    )

        @given(hostname_labels(allow_idn=False))
        def test_hostname_labels_valid_ascii(self, label):
            # type: (Text) -> None
            """
            hostname_labels() generates a ASCII host name labels.
            """
            try:
                check_label(label)
                label.encode("ascii")
            except UnicodeError:  # pragma: no cover
                raise AssertionError("Invalid ASCII label: {!r}".format(label))

        @given(hostnames())
        def test_hostnames_idn(self, hostname):
            # type: (Text) -> None
            """
            hostnames() generates a IDN host names.
            """
            try:
                for label in hostname.split(u"."):
                    check_label(label)
                idna_encode(hostname)
            except UnicodeError:  # pragma: no cover
                raise AssertionError(
                    "Invalid IDN host name: {!r}".format(hostname)
                )

        @given(hostnames(allow_leading_digit=False))
        def test_hostnames_idn_nolead(self, hostname):
            # type: (Text) -> None
            """
            hostnames(allow_leading_digit=False) generates a IDN host names
            without leading digits.
            """
            self.assertTrue(hostname == hostname.lstrip(digits))

        @given(hostnames(allow_idn=False))
        def test_hostnames_ascii(self, hostname):
            # type: (Text) -> None
            """
            hostnames() generates a ASCII host names.
            """
            try:
                for label in hostname.split(u"."):
                    check_label(label)
                hostname.encode("ascii")
            except UnicodeError:  # pragma: no cover
                raise AssertionError(
                    "Invalid ASCII host name: {!r}".format(hostname)
                )

        @given(hostnames(allow_leading_digit=False, allow_idn=False))
        def test_hostnames_ascii_nolead(self, hostname):
            # type: (Text) -> None
            """
            hostnames(allow_leading_digit=False, allow_idn=False) generates
            ASCII host names without leading digits.
            """
            self.assertTrue(hostname == hostname.lstrip(digits))

        @given(paths())
        def test_paths(self, path):
            # type: (Sequence[Text]) -> None
            """
            paths() generates sequences of URL path components.
            """
            text = u"/".join(path)
            try:
                text.encode("utf-8")
            except UnicodeError:  # pragma: no cover
                raise AssertionError("Invalid URL path: {!r}".format(path))

            for segment in path:
                self.assertNotIn("#/?", segment)

        @given(encoded_urls())
        def test_encoded_urls(self, url):
            # type: (EncodedURL) -> None
            """
            encoded_urls() generates EncodedURLs.
            """
            self.assertIsInstance(url, EncodedURL)

        @given(decoded_urls())
        def test_decoded_urls(self, url):
            # type: (DecodedURL) -> None
            """
            decoded_urls() generates DecodedURLs.
            """
            self.assertIsInstance(url, DecodedURL)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from typing import cast


from .. import _url
from .common import HyperlinkTestCase
from .._url import register_scheme, URL, DecodedURL


class TestSchemeRegistration(HyperlinkTestCase):
    def setUp(self):
        # type: () -> None
        self._orig_scheme_port_map = dict(_url.SCHEME_PORT_MAP)
        self._orig_no_netloc_schemes = set(_url.NO_NETLOC_SCHEMES)

    def tearDown(self):
        # type: () -> None
        _url.SCHEME_PORT_MAP = self._orig_scheme_port_map
        _url.NO_NETLOC_SCHEMES = self._orig_no_netloc_schemes

    def test_register_scheme_basic(self):
        # type: () -> None
        register_scheme("deltron", uses_netloc=True, default_port=3030)

        u1 = URL.from_text("deltron://example.com")
        assert u1.scheme == "deltron"
        assert u1.port == 3030
        assert u1.uses_netloc is True

        # test netloc works even when the original gives no indication
        u2 = URL.from_text("deltron:")
        u2 = u2.replace(host="example.com")
        assert u2.to_text() == "deltron://example.com"

        # test default port means no emission
        u3 = URL.from_text("deltron://example.com:3030")
        assert u3.to_text() == "deltron://example.com"

        register_scheme("nonetron", default_port=3031)
        u4 = URL(scheme="nonetron")
        u4 = u4.replace(host="example.com")
        assert u4.to_text() == "nonetron://example.com"

    def test_register_no_netloc_scheme(self):
        # type: () -> None
        register_scheme("noloctron", uses_netloc=False)
        u4 = URL(scheme="noloctron")
        u4 = u4.replace(path=("example", "path"))
        assert u4.to_text() == "noloctron:example/path"

    def test_register_no_netloc_with_port(self):
        # type: () -> None
        with self.assertRaises(ValueError):
            register_scheme("badnetlocless", uses_netloc=False, default_port=7)

    def test_invalid_uses_netloc(self):
        # type: () -> None
        with self.assertRaises(ValueError):
            register_scheme("badnetloc", uses_netloc=cast(bool, None))
        with self.assertRaises(ValueError):
            register_scheme("badnetloc", uses_netloc=cast(bool, object()))

    def test_register_invalid_uses_netloc(self):
        # type: () -> None
        with self.assertRaises(ValueError):
            register_scheme("lol", uses_netloc=cast(bool, object()))

    def test_register_invalid_port(self):
        # type: () -> None
        with self.assertRaises(ValueError):
            register_scheme("nope", default_port=cast(bool, object()))

    def test_register_no_quote_plus_scheme(self):
        # type: () -> None
        register_scheme("keepplus", query_plus_is_space=False)
        plus_is_not_space = DecodedURL.from_text(
            "keepplus://example.com/?q=a+b"
        )
        plus_is_space = DecodedURL.from_text("https://example.com/?q=a+b")
        assert plus_is_not_space.get("q") == ["a+b"]
        assert plus_is_space.get("q") == ["a b"]



from ._url import (
    parse,
    register_scheme,
    URL,
    EncodedURL,
    DecodedURL,
    URLParseError,
)

__all__ = (
    "parse",
    "register_scheme",
    "URL",
    "EncodedURL",
    "DecodedURL",
    "URLParseError",
)
